(*
   RE - A regular expression library

   Copyright (C) 2001 Jerome Vouillon
   email: Jerome.Vouillon@inria.fr

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)

(*
To do next
==========
- Posix extended regular extension
    - a* ==> longest ((shortest (no_group a)* ), a | ())  (!!!)
    - abc understood as (ab)c
    - "((a?)|b)" against "ab" should not bind the first subpattern to anything
- Reduce memory usage
    - More compact representation of character sequences
    - Special notation for "anything but this set of characters"
    - Maybe, a way to denote intervals of characters
- Simple optimisations
    - alt containing alt
    - epsilon elimination
    - Seq (Seq (x,y), z) => Seq (x, Seq (y, z)) under some circumptances
      (x or y has a fixed length)
     ...
- Special case for last new_line
- Try to optimize the main loops
- Rewrite the main loops in C
- There must be a way to handle "(((ab)c)d)e" efficiently

Later
=====
- Back-reference
- Look-ahead/look-behind ???
*)

type sem =
  [ `Longest
  | `Shortest
  | `First ]

type rep_kind =
    Greedy
  | Non_greedy

type category = int

type expr =
  { id : int;
    def : def }

and def =
    Cst of int list
  | Alt of expr list
  | Seq of sem * expr * expr
  | Eps
  | Rep of rep_kind * sem * expr
  | Anchor of int
  | Before of category
  | After of category

type e =
    TSeq of e list * expr * sem
  | TExp of (int * int) list * expr
  | TMatch of (int * int) list

(****)

let print_kind ch k =
  Format.fprintf ch "%s"
    (match k with
       `Shortest -> "short"
     | `Longest  -> "long"
     | `First    -> "first")

let rec print_expr ch e =
  match e.def with
    Cst l ->
      Format.fprintf ch "@[<3>(cst";
      List.iter (fun c -> Format.fprintf ch "@ %d" c) l;
      Format.fprintf ch ")@]"
  | Alt l ->
      Format.fprintf ch "@[<3>(alt";
      List.iter (fun e -> Format.fprintf ch "@ %a" print_expr e) l;
      Format.fprintf ch ")@]"
  | Seq (k, e, e') ->
      Format.fprintf ch "@[<3>(seq %a@ %a@ %a)@]"
        print_kind k print_expr e print_expr e'
  | Eps ->
      Format.fprintf ch "eps"
  | Rep (rk, k, e) ->
      Format.fprintf ch "@[<3>(rep@ %a)@]" print_expr e
  | Anchor i ->
      Format.fprintf ch "@[<3>(anc@ %d)@]" i
  | Before c ->
      Format.fprintf ch "@[<3>(before@ %d)@]" c
  | After c ->
      Format.fprintf ch "@[<3>(after@ %d)@]" c

(****)

let rec first f l =
  match l with
    [] ->
      None
  | x :: r ->
      match f x with
        None          -> first f r
      | Some _ as res -> res

let rec repeat n r rem =
  if n = 0 then rem else
  r :: repeat (n - 1) r rem

let rec iter n f v =
  if n = 0 then v else
  iter (n - 1) f (f v)

(****)

let eps = { id = 0; def = Eps }

let mk_expr ids def =
  incr ids;
  { id = !ids; def = def }

let seq ids kind x y =
  match x.def, y.def with
    Alt [], _                 -> mk_expr ids (Alt [])
  | _, Alt []                 -> mk_expr ids (Alt [])
  | Eps, _                    -> y
  | _, Eps when kind = `First -> x
  | _                         -> mk_expr ids (Seq (kind, x, y))

let texp anch x = TExp (anch, x)

let tseq kind x y rem =
  match x with
    []                         -> rem
  | [TExp (anch, {def = Eps})] -> TExp (anch, y) :: rem
  | _                          -> TSeq (x, y, kind) :: rem

(****)

type state =
  { idx : int;
    real_idx : int;
    next : state array;
    mutable final : (int * (int * int) list option) option;
    cat : category;
    desc : e list }

let hash_combine h accu = accu * 65599 + h

let rec hash_anchors l accu =
  match l with
    [] ->
      accu
  | (a, i) :: r ->
      hash_anchors r (hash_combine a (hash_combine i accu))

let rec hash_e l accu =
  match l with
    [] ->
      accu
  | TSeq (l', e, _) :: r ->
      hash_e r (hash_combine 0x172a1bce (hash_combine e.id (hash_e l' accu)))
  | TExp (anch, e) :: r ->
      hash_e r
        (hash_combine 0x2b4c0d77 (hash_combine e.id (hash_anchors anch accu)))
  | TMatch anch :: r ->
      hash_e r (hash_combine 0x1c205ad5 (hash_anchors anch accu))

let hash_state idx cat desc =
  hash_e desc (hash_combine idx (hash_combine cat 0)) land 0x3FFFFFFF

let rec equal_e l1 l2 =
  match l1, l2 with
    [], [] ->
      true
  | TSeq (l1', e1, _) :: r1, TSeq (l2', e2, _) :: r2 ->
      e1.id = e2.id && equal_e l1' l2' && equal_e r1 r2
  | TExp (anch1, e1) :: r1, TExp (anch2, e2) :: r2 ->
      e1.id = e2.id && anch1 = anch2 && equal_e r1 r2
  | TMatch anch1 :: r1, TMatch anch2 :: r2 ->
      anch1 = anch2 && equal_e r1 r2
  | _ ->
      false

let equal_state (idx1, cat1, desc1) (idx2, cat2, desc2) =
  (idx1 : int) = idx2 && (cat1 : int) = cat2 && equal_e desc1 desc2

module States =
  Hashtbl.Make
    (struct
       type t = (int * category * e list) * int
       let equal (x, h) (x', h') = (h : int) = h' && equal_state x x'
       let hash (x, h) = h
     end)

type rx =
  { initial : expr;
    mutable initial_states : (int * state) list;
    cols : string;
    col_repr : string;
    ncol : int;
    mutable tbl : bool array;
    states : state States.t }

type info =
  { rx : rx;
    i_cols : string;
    mutable positions : int array;
    mutable pos : int;
    last : int }

(****)

let cat_inexistant = 1
let cat_letter = 2
let cat_not_letter = 4
let cat_newline = 8
let cat_lastnewline = 16
let cat_start = 32

let category rx c =
  if c = -1 then cat_inexistant else
  match rx.col_repr.[c] with
    'a'..'z' | 'A'..'Z' ->
      cat_letter
  | '\n' ->
      cat_not_letter lor cat_newline
  | _ ->
      cat_not_letter

(**** Find a free index ****)

let reset_table a = Array.fill a 0 (Array.length a) false

let rec mark_used_indices tbl l =
  List.iter
    (fun x ->
       match x with
         TSeq (l, _, _) ->
           mark_used_indices tbl l
       | TExp (anch, _) ->
           List.iter (fun (_, i) -> if i >= 0 then tbl.(i) <- true) anch
       | TMatch anch ->
           List.iter (fun (_, i) -> if i >= 0 then tbl.(i) <- true) anch)
    l
  
let rec find_free tbl idx len =
  if idx = len || not tbl.(idx) then idx else find_free tbl (idx + 1) len

let free_index info l =
  let rx = info.rx in
  reset_table rx.tbl;
  mark_used_indices rx.tbl l;
  let len = Array.length rx.tbl in
  let idx = find_free rx.tbl 0 len in
  if idx = len then begin
    rx.tbl <- Array.make (2 * len) false;
    let pos = info.positions in
    if Array.length info.positions > 0 then begin
      info.positions <- Array.make (2 * len) 0;
      Array.blit pos 0 info.positions 0 len
    end
  end;
  idx

(**** Computation of the next state ****)

let remove_matches l =
  List.filter (fun x -> match x with TMatch _ -> false | _ -> true) l

let rec split_at_match_rec l' l =
  match l with
    []                 -> assert false
  | TMatch _ as x :: r -> (List.rev l', remove_matches r)
  | x :: r             -> split_at_match_rec (x :: l') r

let split_at_match l = split_at_match_rec [] l

let print_anch ch l =
  match l with
    [] ->
      ()
  | (a, i) :: r ->
      Format.fprintf ch "%d-%d" a i;
      List.iter (fun (a, i) -> Format.fprintf ch " %d-%d" a i) r

let rec print_state_rec ch l y =
  match l with
    [] ->
      ()
  | TMatch anch as x :: r ->
      Format.fprintf ch "match (%a)" print_anch anch
  | TSeq (l', x, kind) :: r ->
      print_state_rec ch l' x;
      print_state_rec ch r y;
  | TExp (anch, {def = Eps}) as e :: r ->
      Format.fprintf ch "%d (%a) " y.id print_anch anch;
      print_state_rec ch r y
  | TExp (anch, x) as e :: r ->
      Format.fprintf ch "%d (%a) " x.id print_anch anch;
      print_state_rec ch r y
let print_state ch l = print_state_rec ch l eps

let rec remove_duplicates prev l y =
  match l with
    [] ->
      ([], prev)
  | TMatch _ as x :: r -> (* Truncate after first match *)
      ([x], prev)
  | TSeq (l', x, kind) :: r ->
      let (l'', prev') = remove_duplicates prev l' x in
      let (r', prev'') = remove_duplicates prev' r y in
      (tseq kind l'' x r', prev'')
  | TExp (anch, {def = Eps}) as e :: r ->
      if List.memq y.id prev then
        remove_duplicates prev r y
      else
        let (r', prev') = remove_duplicates (y.id :: prev) r y in
        (e :: r', prev')
  | TExp (anch, x) as e :: r ->
      if List.memq x.id prev then
        remove_duplicates prev r y
      else
        let (r', prev') = remove_duplicates (x.id :: prev) r y in
        (e :: r', prev')

let rec anch_set_idx used idx anch =
  match anch with
    (a, -1) :: rem ->
      used := true;
      (a, idx) :: anch_set_idx used idx rem
  | _ ->
      anch

let rec set_idx used idx l =
  match l with
    [] ->
      []
  | TMatch anch :: r ->
      TMatch (anch_set_idx used idx anch) :: set_idx used idx r
  | TSeq (l', x, kind) :: r ->
      TSeq (set_idx used idx l', x, kind) :: set_idx used idx r
  | TExp (anch, x) :: r ->
      TExp (anch_set_idx used idx anch, x) :: set_idx used idx r

let count =ref 0
let rec delta_1 anch c cat' cat x rem =
(*incr count; Format.eprintf ">> %d@." !count;*)
  match x.def with
    Cst s ->
      if List.memq c s then texp anch eps :: rem else rem
  | Alt l ->
      delta_2 anch c cat' cat l rem
  | Seq (kind, y, z) ->
      let y' = delta_1 anch c cat' cat y [] in
      delta_seq c cat' cat kind y' z rem
  | Rep (rep_kind, kind, y) ->
      let y' = delta_1 anch c cat' cat y [] in
      let (y'', anch') =
        match
          first (fun x -> match x with TMatch anch -> Some anch | _ -> None) y'
        with
          None       -> (y', anch)
        | Some anch' -> (remove_matches y', anch')
      in
      begin match rep_kind with
        Greedy     -> tseq kind y'' x (TMatch anch' :: rem)
      | Non_greedy -> TMatch anch :: tseq kind y'' x rem
      end
  | Eps ->
      TMatch anch :: rem
  | Anchor i ->
      TMatch ((i, -1) :: List.remove_assq i anch) :: rem
  | Before cat'' ->
      if cat land cat'' <> 0 then TMatch anch :: rem else rem
  | After cat'' ->
      if cat' land cat'' <> 0 then TMatch anch :: rem else rem

and delta_2 anch c cat' cat l rem =
  match l with
    []     -> rem
  | y :: r -> delta_1 anch c cat' cat y (delta_2 anch c cat' cat r rem)

and delta_seq c cat' cat kind y z rem =
  match
    first (fun x -> match x with TMatch anch -> Some anch | _ -> None) y
  with
    None ->
      tseq kind y z rem
  | Some anch ->
      match kind with
        `Longest ->
          tseq kind (remove_matches y) z (delta_1 anch c cat' cat z rem)
      | `Shortest ->
          delta_1 anch c cat' cat z (tseq kind (remove_matches y) z rem)
      | `First ->
          let (y', y'') = split_at_match y in
          tseq kind y' z (delta_1 anch c cat' cat z (tseq kind y'' z rem))

let rec delta_3 c cat' cat x rem =
  match x with
    TSeq (y, z, kind) ->
      let y' = delta_4 c cat' cat y [] in
      delta_seq c cat' cat kind y' z rem
  | TExp (anch, e) ->
      delta_1 anch c cat' cat e rem
  | TMatch _ ->
      x :: rem

and delta_4 c cat' cat l rem =
  match l with
    []     -> rem
  | y :: r -> delta_3 c cat' cat y (delta_4 c cat' cat r rem)

let unknown = -2
let abort = -3

let count = ref 0
let delta info ~cat ~char ~expr =
(*incr count; Format.eprintf "> %d@." !count;*)
  let cat' = category info.rx char in
  let (expr', _) =
    remove_duplicates [] (delta_4 char cat cat' expr []) eps in
  let idx = free_index info expr' in
  let used = ref false in
  let expr'' = set_idx used idx expr' in
  (idx, cat', expr'')

let dummy_next = [||]

let unknown_state =
  { idx = unknown; real_idx = 0;
    next = dummy_next; final = None;
    cat = 0; desc = [TExp ([], eps)] }

let count = ref 0
let mk_state ncol idx cat desc =
(*Format.eprintf "%a -- %d@." print_state desc idx;*)
(*incr count; Format.eprintf "%d@." !count;*)
(*Format.eprintf "%d %x@."cat (Hashtbl.hash_param 10000 1000 desc);*)
  let abort_state =
    match desc with
      [] | TMatch _ :: _ ->
        true
    | _ ->
        false
  in
  { idx = if abort_state then abort else idx;
    real_idx = idx;
    next = if abort_state then dummy_next else Array.make ncol unknown_state;
    final = None;
    cat = cat; desc = desc }

let find_state rx idx cat desc =
  let key = ((idx, cat, desc), hash_state idx cat desc) in
  try
    States.find rx.states key
  with Not_found ->
    let st = mk_state rx.ncol idx cat desc in
    States.add rx.states key st;
    st

(**** Match with anchors ****)

(*
let rec loop info s pos st =
  if pos < info.last then
    let st' = st.next.(Char.code info.i_cols.[Char.code s.[pos]]) in
    let idx = st'.idx in
    if idx >= 0 then begin
      info.pos.(idx) <- pos;
      loop info s (pos + 1) st'
    end else if idx = abort then begin
      info.pos.(st'.real_idx) <- pos;
      st'
    end else (* Unknown *)
      validate info s pos st
  else
    st
*)

let rec loop info s pos st =
  if pos < info.last then
    let st' = st.next.(Char.code info.i_cols.[Char.code s.[pos]]) in
    loop2 info s pos st st'
  else
    st

and loop2 info s pos st st' =
  let idx = st'.idx in
  if idx >= 0 then begin
    let pos = pos + 1 in
    if pos < info.last then begin
      (* It is important to place these reads before the write *)
      (* But then, we don't have enough registers left to store the
         right position.  So, we store the position plus one. *)
      let st'' = st'.next.(Char.code info.i_cols.[Char.code s.[pos]]) in
      info.positions.(idx) <- pos;
      loop2 info s pos st' st''
    end else begin
      info.positions.(idx) <- pos;
      st'
    end
  end else if idx = abort then begin
    info.positions.(st'.real_idx) <- pos + 1;
    st'
  end else (* Unknown *)
    validate info s pos st

and validate info s pos st =
  let c = Char.code info.i_cols.[Char.code s.[pos]] in
  let (idx, cat, e') = delta info st.cat c st.desc in
  let st' = find_state info.rx idx cat e' in
  st.next.(c) <- st';
  loop info s pos st

let rec loop_no_anchor info s pos last st =
  if pos < last then
    let st' = st.next.(Char.code info.i_cols.[Char.code s.[pos]]) in
    let idx = st'.idx in
    if idx >= 0 then
      loop_no_anchor info s (pos + 1) last st'
    else if idx = abort then
      st'
    else (* Unknown *)
      validate info s pos last st
  else
    st

and validate info s pos last st =
  let c = Char.code info.i_cols.[Char.code s.[pos]] in
  let (idx, cat, e') = delta info st.cat c st.desc in
  let st' = find_state info.rx idx cat e' in
  st.next.(c) <- st';
  loop_no_anchor info s pos last st

let final info st =
  match st.final with
    Some v -> v
  | None ->
      let (idx, _, e) = delta info st.cat (-1) st.desc in
      let res =
        first (fun x -> match x with TMatch anc -> Some anc | _ -> None) e in
      st.final <- Some (idx, res);
      (idx, res)

let find_initial_state rx cat =
  try
    List.assq cat rx.initial_states
  with Not_found ->
    let st = find_state rx 0 cat [TExp ([], rx.initial)] in
    rx.initial_states <- (cat, st) :: rx.initial_states;
    st

let match_str groups rx s pos =
  let info =
    { rx = rx; i_cols = rx.cols; pos = pos; last = String.length s;
      positions = if groups then Array.make (Array.length rx.tbl) 0 else [||] }
  in
  let initial_cat =
    cat_start lor
    if pos = 0 then cat_inexistant else
    category rx (Char.code rx.cols.[Char.code s.[pos - 1]])
  in
  let initial_state = find_initial_state rx initial_cat in
  let st =
    if groups then
      loop info s 0 initial_state
    else 
      loop_no_anchor info s 0 (String.length s) initial_state
  in
  if st.idx = abort then
    match st.desc with
      []              -> None
    | TMatch anc :: _ -> Some (s, anc, info.positions)
    | _               -> assert false
  else
    let (idx, res) = final info st in
    info.positions.(idx) <- String.length s + 1;
    match res with
      None     -> None
    | Some anc -> if groups then Some (s, anc, info.positions)
                  else Some (s, [], info.positions)

let mk_rx init cols col_repr ncol =
  { initial = init;
    initial_states = [];
    cols = cols;
    col_repr = col_repr;
    ncol = ncol;
    tbl = [| false |];
    states = States.create 97 }

(**** Character sets ****)

type cset = (int * int) list

let rec cunion l l' =
  match l, l' with
    _, [] -> l
  | [], _ -> l'
  | (c1, c2)::r, (c1', c2')::r' ->
      if c2 + 1 < c1' then
        (c1, c2)::cunion r l'
      else if c2' + 1 < c1 then
        (c1', c2')::cunion l r'
      else if c2 < c2' then
        cunion r ((min c1 c1', c2')::r')
      else
        cunion ((min c1 c1', c2)::r) r'

let rec cinter l l' =
  match l, l' with
    _, [] -> []
  | [], _ -> []
  | (c1, c2)::r, (c1', c2')::r' ->
      if c2 < c1' then
        cinter r l'
      else if c2' < c1 then
        cinter l r'
      else if c2 < c2' then
        (max c1 c1', c2)::cinter r l'
      else
        (max c1 c1', c2')::cinter l r'

let rec cnegate mi ma l =
  match l with
    [] ->
      if mi <= ma then [(mi, ma)] else []
  | (c1, c2)::r when ma < c1 ->
      if mi <= ma then [(mi, ma)] else []
  | (c1, c2)::r when mi < c1 ->
      (mi, c1 - 1) :: cnegate c1 ma l
  | (c1, c2)::r (* when c1 <= mi *) ->
      cnegate (max mi (c2 + 1)) ma r

let csingle c = let i = Char.code c in [i, i]

let cadd c l = cunion (csingle c) l

let cseq c c' =
  let i = Char.code c in let i' = Char.code c' in
  if i <= i' then [i, i'] else [i', i]

let rec ctrans o l =
  match l with
    [] -> []
  | (c1, c2) :: r ->
      if c2 + o < 0 || c1 + o > 255 then
        ctrans o r
      else
        (c1 + o, c2 + o) :: ctrans o r

let cempty = []

let cany = [0, 255]

let rec interval i j = if i > j then [] else i :: interval (i + 1) j

let rec cset_hash_rec l =
  match l with
    []        -> 0
  | (i, j)::r -> i + 13 * j + 257 * cset_hash_rec r
let cset_hash l = (cset_hash_rec l) land 0x3FFFFFFF

module CSetMap =
  Map.Make
  (struct
    type t = int * (int * int) list
    let compare (i, u) (j, v) =
      let c = compare i j in if c <> 0 then c else compare u v
   end)

let trans_set cache cm s =
  match s with
    [i, j] when i = j ->
      [Char.code cm.[i]] 
  | _ ->
      let v = (cset_hash_rec s, s) in
      try
        CSetMap.find v !cache
      with Not_found ->
        let l =
          List.fold_right
            (fun (i, j) l -> cunion [Char.code cm.[i], Char.code cm.[j]] l)
            s []
        in
        let res =
          List.flatten (List.map (fun (i, j) -> interval i j) l)
        in
        cache := CSetMap.add v res !cache;
        res

(****)

type sem_status = Compulsory | Indicative

type regexp =
    Set of cset
  | Sequence of regexp list
  | Alternative of regexp list
  | Repeat of regexp * int * int option
  | Beg_of_line | End_of_line
  | Beg_of_word | End_of_word | Not_bound
  | Beg_of_str | End_of_str
  | Last_end_of_line | Start
  | Sem of sem_status * sem * regexp
  | Sem_greedy of rep_kind * regexp
  | Group of regexp | No_group of regexp
  | Case of regexp | No_case of regexp
  | Intersection of regexp list
  | Complement of regexp list
  | Difference of regexp * regexp

let rec is_charset r =
  match r with
    Set _ ->
      true
  | Alternative l | Intersection l | Complement l ->
      List.for_all is_charset l
  | Difference (r, r') ->
      is_charset r && is_charset r'
  | Sem (_, _, r) | Sem_greedy (_, r)
  | No_group r | Case r | No_case r ->
      is_charset r
  | Sequence _ | Repeat _ | Beg_of_line | End_of_line
  | Beg_of_word | End_of_word | Beg_of_str | End_of_str
  | Not_bound | Last_end_of_line | Start | Group _ ->
      false

(**** Colormap ****)

(*XXX Use a better algorithm allowing non-contiguous regions? *)
let rec split s cm =
  match s with
    []    -> ()
  | (i, j)::r -> cm.[i] <- '\001'; cm.[j + 1] <- '\001'; split r cm

let cupper =
  cunion (cseq 'A' 'Z') (cunion (cseq '\192' '\214') (cseq '\216' '\222'))
let clower = ctrans 32 cupper
let calpha = cadd '\170' (cadd '\186' (cunion clower cupper))
let cdigit = cseq '0' '9'
let calnum = cunion calpha cdigit
let cword = cadd '_' calnum

let rec colorize c regexp =
  let rec colorize regexp =
    match regexp with
      Set s                     -> split s c
    | Sequence l                -> List.iter colorize l
    | Alternative l             -> List.iter colorize l
    | Repeat (r, _, _)          -> colorize r
    | Beg_of_line | End_of_line -> split (csingle '\n') c
    | Beg_of_word | End_of_word -> split cword c
    | Beg_of_str | End_of_str   -> ()
    | Start                     -> ()
    | Last_end_of_line          -> split (csingle '\n') c
    | Sem (_, _, r)             -> colorize r
    | Group r | No_group r      -> colorize r
    | Case _ | _
    | Intersection _
    | Complement _
    | Difference _              -> assert false
  in
  colorize regexp

let make_cmap () = String.make 257 '\000'

let flatten_cmap cm =
  let c = String.create 256 in
  let col_repr = String.create 256 in
  let v = ref 0 in
  c.[0] <- '\000';
  col_repr.[0] <- '\000';
  for i = 1 to 255 do
    if cm.[i] <> '\000' then incr v;
    c.[i] <- Char.chr !v;
    col_repr.[!v] <- Char.chr i
  done;
  (c, String.sub col_repr 0 (!v + 1), !v + 1)

(**** Compilation ****)

let sequence l =
  match l with
    [x] -> x
  | l   -> Sequence l

let rec merge_sequences l =
  match l with
    [] ->
      l
  | Alternative l' :: r ->
      merge_sequences (l' @ r)
  | Sequence (x :: y) :: r ->
      begin match merge_sequences r with
        Sequence (x' :: y') :: r' when x = x' ->
          Sequence [x; Alternative [sequence y; sequence y']] :: r'
      | r' ->
          Sequence (x :: y) :: r'
      end
  | x :: r ->
      x :: merge_sequences r

let rec rename ids x =
  match x.def with
    Cst _ | Eps | Anchor _ | Before _ | After _ ->
      mk_expr ids x.def
  | Alt l ->
      mk_expr ids (Alt (List.map (rename ids) l))
  | Seq (k, y, z) ->
      mk_expr ids (Seq (k, rename ids y, rename ids z))
  | Rep (g, k, y) ->
      mk_expr ids (Rep (g, k, rename ids y))

let enforce_kind ids kind kind' cr =
  match kind, kind' with
    `First, `First -> cr
  | `First,  k     -> seq ids k cr (mk_expr ids Eps)
  |  _             -> cr

(* XXX should probably compute a category mask *)
let rec translate ids kind ign_group ign_case greedy pos cache c r =
  match r with
    Set s ->
      (begin match trans_set cache c s with
         [] -> mk_expr ids (Alt []) (* XXX??? *)
       | l  -> mk_expr ids (Cst l)
       end,
       kind)
  | Sequence l ->
      (trans_seq ids kind ign_group ign_case greedy pos cache c l, kind)
  | Alternative l ->
      begin match merge_sequences l with
        [r'] ->
          let (cr, kind') =
            translate ids kind ign_group ign_case greedy pos cache c r' in
          (enforce_kind ids kind kind' cr, kind)
      | l' ->
          (mk_expr ids
             (Alt
                (List.map
                   (fun r' ->
                      let (cr, kind') =
                        translate ids kind ign_group ign_case greedy
                          pos cache c r' in
                      enforce_kind ids kind kind' cr)
                   (merge_sequences l))),
           kind)
      end
  | Repeat (r', i, j) ->
      let (cr, kind') =
        translate ids kind ign_group ign_case greedy pos cache c r' in
      let rem =
        match j with
          None ->
            mk_expr ids (Rep (greedy, kind', cr))
        | Some j ->
            let f =
              match greedy with
                Greedy ->
                  fun rem ->
                    mk_expr ids (Alt [seq ids kind' (rename ids cr) rem;
                                      mk_expr ids Eps])
              | Non_greedy ->
                  fun rem ->
                    mk_expr ids (Alt [mk_expr ids Eps;
                                      seq ids kind' (rename ids cr) rem])
            in
            iter (j - i) f (mk_expr ids Eps)
      in
      (iter i (fun rem -> seq ids kind' (rename ids cr) rem) rem, kind)
  | Beg_of_line ->
      (mk_expr ids (After (cat_inexistant lor cat_newline)), kind)
  | End_of_line ->
      (mk_expr ids (Before (cat_inexistant lor cat_newline)), kind)
  | Beg_of_word ->
      (seq ids `First
           (mk_expr ids (After (cat_inexistant lor cat_not_letter)))
           (mk_expr ids (Before (cat_inexistant lor cat_letter))),
       kind)
  | End_of_word ->
      (seq ids `First
           (mk_expr ids (After (cat_inexistant lor cat_letter)))
           (mk_expr ids (Before (cat_inexistant lor cat_not_letter))),
       kind)
  | Not_bound ->
      (mk_expr ids
         (Alt [seq ids `First
                 (mk_expr ids (After cat_letter))
                 (mk_expr ids (Before cat_letter));
               seq ids `First
                 (mk_expr ids (After cat_letter))
                 (mk_expr ids (Before cat_letter))]),
       kind)
  | Beg_of_str ->
      (mk_expr ids (After cat_inexistant), kind)
  | End_of_str ->
      (mk_expr ids (Before cat_inexistant), kind)
  | Last_end_of_line ->
      (mk_expr ids (Before (cat_inexistant lor cat_lastnewline)), kind)
  | Start ->
      (mk_expr ids (After cat_start), kind)
  | Sem (Compulsory, kind', r') ->
      let (cr, kind'') =
        translate ids kind' ign_group ign_case greedy pos cache c r' in
      (enforce_kind ids kind' kind'' cr,
       kind')
  | Sem (Indicative, kind', r') ->
      translate ids kind' ign_group ign_case greedy pos cache c r'
  | Sem_greedy (greedy, r') ->
      translate ids kind ign_group ign_case greedy pos cache c r'
  | Group r' ->
      if ign_group then 
        translate ids kind ign_group ign_case greedy pos cache c r'
      else
        let p = !pos in
        pos := !pos + 2;
        let (cr, kind') =
          translate ids kind ign_group ign_case greedy pos cache c r' in
        (seq ids `First (mk_expr ids (Anchor p)) (
         seq ids `First cr (mk_expr ids (Anchor (p + 1)))),
         kind')
  | No_group r' ->
      translate ids kind true ign_case greedy pos cache c r'
  | Difference _ | Complement _ | Intersection _ | No_case _ | Case _ ->
      assert false

and trans_seq ids kind ign_group ign_case greedy pos cache c l =
  match l with
    [] ->
      mk_expr ids Eps
  | [r] ->
      let (cr', kind') =
        translate ids kind ign_group ign_case greedy pos cache c r in
      enforce_kind ids kind kind' cr'
  | r :: rem ->
      let (cr', kind') =
        translate ids kind ign_group ign_case greedy pos cache c r in
      let cr'' =
        trans_seq ids kind ign_group ign_case greedy pos cache c rem in
      if cr''.def = Eps then
        cr'
      else if cr'.def = Eps then
        cr''
      else
        seq ids kind' cr' cr''

(**** Case ****)

let case_insens s =
  cunion s (cunion (ctrans 32 (cinter s cupper))
                   (ctrans (-32) (cinter s clower)))

let as_set r =
  match r with
    Set s -> s
  | _     -> assert false

(* XXX Should split alternatives into (1) charsets and (2) more
   complex regular expressions; alternative should therefore probably
   be flatten here *)
let rec handle_case ign_case r =
  match r with
    Set s ->
      Set (if ign_case then case_insens s else s)
  | Sequence l ->
      Sequence (List.map (handle_case ign_case) l)
  | Alternative l ->
      let l' = List.map (handle_case ign_case) l in
      if is_charset (Alternative l') then
        Set (List.fold_left (fun s r -> cunion s (as_set r)) cempty l)
      else
        Alternative l'
  | Repeat (r, i, j) ->
      Repeat (handle_case ign_case r, i, j)
  | Beg_of_line | End_of_line | Beg_of_word | End_of_word | Not_bound
  | Beg_of_str | End_of_str | Last_end_of_line | Start ->
      r
  | Sem (status, k, r) ->
      let r' = handle_case ign_case r in
      if is_charset r' then r' else
      Sem (status, k, r')
  | Sem_greedy (k, r) ->
      Sem_greedy (k, handle_case ign_case r)
  | Group r ->
      Group (handle_case ign_case r)
  | No_group r ->
      let r' = handle_case ign_case r in
      if is_charset r' then r' else
      No_group r'
  | Case r ->
      handle_case false r
  | No_case r ->
      handle_case true r
  | Intersection l ->
      Set (List.fold_left (fun s r -> cinter s (as_set r)) cany l)
  | Complement l ->
      Set (cnegate 0 255
             (List.fold_left (fun s r -> cunion s (as_set r)) cempty l))
  | Difference (r, r') ->
      Set (cinter (as_set r) (cnegate 0 255 (as_set r')))

(****)

let compile_1 regexp =
  let regexp = handle_case false regexp in
  let c = make_cmap () in
  colorize c regexp;
  let (col, col_repr, ncol) = flatten_cmap c in
  let ids = ref 0 in
  let pos = ref 0 in
  let (r, kind) =
    translate ids
      `First false false Greedy pos (ref CSetMap.empty) col regexp in
  let r = enforce_kind ids `First kind r in
(*Format.eprintf "<%d %d>@." !ids ncol;*)
  mk_rx r col col_repr ncol

(****)

type t = regexp

let str s =
  let l = ref [] in
  for i = String.length s - 1 downto 0 do
    l := Set (csingle s.[i]) :: !l
  done;
  Sequence !l
let char c = Set (csingle c)

let alt l = Alternative l
let seq l = Sequence l
let empty = alt []
let epsilon = seq []
let repn r i j =
  if i < 0 then invalid_arg "Rx.rep";
  begin match j with Some j when j < i -> invalid_arg "Rx.rep" | _ -> () end;
  Repeat (r, i, j)
let rep r = repn r 0 None
let rep1 r = repn r 1 None
let opt r = repn r 0 (Some 1)
let bol = Beg_of_line
let eol = End_of_line
let bow = Beg_of_word
let eow = End_of_word
let word r = seq [bow; r; eow]
let not_boundary = Not_bound
let bos = Beg_of_str
let eos = End_of_str
let leol = Last_end_of_line
let start = Start
let longest r = Sem (Compulsory, `Longest, r)
let shortest r = Sem (Compulsory, `Shortest, r)
let first r = Sem (Compulsory, `First, r)
let default_sem sem r = Sem (Indicative, sem, r)
let greedy r = Sem_greedy (Greedy, r)
let non_greedy r = Sem_greedy (Non_greedy, r)
let group r = Group r
let no_group r = No_group r

let set str =
  let s = ref [] in
  for i = 0 to String.length str - 1 do
    s := cunion (csingle str.[i]) !s
  done;
  Set !s

let rg c c' = Set (cseq c c')

let inter l =
  let r = Intersection l in
  if is_charset r then r else
  invalid_arg "Rx.inter"

let compl l =
  let r = Complement l in
  if is_charset r then r else
  invalid_arg "Rx.compl"

let diff r r' =
  let r'' = Difference (r, r') in
  if is_charset r'' then r'' else
  invalid_arg "Rx.compl"

let any = Set cany
let notnl = Set (cnegate 0 255 (csingle '\n'))

let lower = alt [rg 'a' 'z'; char '\181'; rg '\223' '\246'; rg '\248' '\255']
let upper = alt [rg 'A' 'Z'; rg '\192' '\214'; rg '\216' '\222']
let alpha = alt [lower; upper; char '\170'; char '\186']
let digit = rg '0' '9'
let alnum = alt [alpha; digit]
let ascii = rg '\000' '\127'
let blank = set "\t "
let cntrl = alt [rg '\000' '\031'; rg '\127' '\159']
let graph = alt [rg '\033' '\126'; rg '\160' '\255']
let print = alt [rg '\032' '\126'; rg '\160' '\255']
let punct =
  alt [rg '\033' '\047'; rg '\058' '\064'; rg '\091' '\096';
       rg '\123' '\126'; rg '\160' '\169'; rg '\171' '\180';
       rg '\182' '\185'; rg '\187' '\191'; char '\215'; char '\247']
let space = alt [char ' '; rg '\009' '\013']
let xdigit = alt [digit; rg 'a' 'f'; rg 'A' 'Z']

let case r = Case r
let no_case r = No_case r

(****)

type substrings = (string * (int * int) list * int array)

let compile ?(sem = `Longest) r =
  compile_1 (seq [shortest (rep any); default_sem sem (group r)])

let exec ?(group = true) ?(pos = 0) rx s =
  match match_str group rx s pos with
    Some substr -> substr
  | None        -> raise Not_found

let rec find_anchor (i : int) l =
  match l with
    [] ->
      raise Not_found
  | (j, idx) :: r ->
      if i = j then idx else find_anchor i r

let get_substring_ofs (s, anch, pos) i =
  (pos.(find_anchor (2 * i) anch) - 1, pos.(find_anchor (2 * i + 1) anch) - 1)
  
let get_substring (s, anch, pos) i =
  let p1 = pos.(find_anchor (2 * i) anch) - 1 in
  let p2 = pos.(find_anchor (2 * i + 1) anch) - 1 in
  String.sub s p1 (p2 - p1)

(**********************************)

(*
Information about the previous character:
- does not exists
- is a letter
- is not a letter
- is a newline
- is last newline

Beginning of word:
- previous is not a letter or does not exist
- current is a letter or does not exist

End of word:
- previous is a letter or does not exist
- current is not a letter or does not exist

Beginning of line:
- previous is a newline or does not exist

Beginning of buffer:
- previous does not exist

End of buffer
- current does not exist

End of line
- current is a newline or does not exist
*)

(*
Rep: e = T,e | ()
  - semantics of the comma (shortest/longest/first)
  - semantics of the union (greedy/non-greedy)

Bounded repetition
  a{0,3} = (a,(a,a?)?)?
*)
